﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Management.Automation;
using System.Management.Automation.Language;
using System.Management.Automation.Runspaces;
using System.Net;
using System.Security.Principal;
using System.Threading.Tasks;

using MvvmHelpers;
using SMM.Helper;


namespace SMM.Automation
{
    public class SimpleScriptRunner : BaseViewModel
    {
        #region fields

        private object instanceLock = new object();
        private PowerShell currentPowerShell;
        private List<(string, object)> proxyVariables = new List<(string, object)>();
        
        #endregion  // fields

        #region constructor

        public SimpleScriptRunner()
        {
            currentPowerShell = PowerShell.Create();
        }

        /// <summary>
        /// Constructs a script runner that can be used for execution of simple PowerShell fragements.
        /// </summary>
        /// <param name="ScriptSource"></param>
        public SimpleScriptRunner(string ScriptSource) {
            currentPowerShell = PowerShell.Create();
            ScriptFragments.Add(ScriptBlock.Create(ScriptSource));
        }

        ~SimpleScriptRunner() {
            // Dispose the PowerShell object and set currentPowerShell to null.
            // It is locked because currentPowerShell may be accessed by the
            // ctrl-C handler.
            lock (instanceLock) {
                currentPowerShell.Runspace?.Dispose();
                currentPowerShell?.Dispose();
                currentPowerShell = null;
            }
        }

        #endregion  // constructor

        #region properties

        public int ExitCode { get; private set; } = 0;

        public bool HadErrors => currentPowerShell?.HadErrors ?? false;

        public Dictionary<string, object> Parameters { get; set; } = new Dictionary<string, object>();

        private PSDataCollection<PSObject> _results = new PSDataCollection<PSObject>();
        public PSDataCollection<PSObject> Results { get => _results; private set => SetProperty(ref _results, value, nameof(Results)); }

        public ObservableCollection<string> Console { get; set; } = new ObservableCollection<string>();

        public ScriptBlock Script { get => ScriptFragments.FirstOrDefault(); }

        public NetworkCredential Credential { get; set; }

        private ObservableCollection<ScriptBlock> _scriptFragments = new ObservableCollection<ScriptBlock>();
        public ObservableCollection<ScriptBlock> ScriptFragments { 
            get => _scriptFragments; 
            set { lock (instanceLock) { _scriptFragments = value; } } 
        }

        public bool IsRunning { get => _isRunning; set { _isRunning = value; OnPropertyChanged(nameof(IsRunning)); } }

        #endregion  // properties

        #region methods

        private bool CheckScript() {
            try
            {
                string _source;
                foreach (var s in ScriptFragments)
                {
                    _source = s.ToString();
                    ScriptBlock.Create(_source);
                }
                return true;
            }
            catch
            {
                return false;
            }
        }

        public ParamBlockAst GetParamBlock() {
            CheckScript();
            dynamic ast = Script?.Ast;
            return ast?.ParamBlock;
        }

        public ScriptRequirements GetScriptRequirements() {
            CheckScript();
            dynamic ast = Script?.Ast;
            return ast?.ScriptRequirements;
        }

        public void SetProxyVariable(string VariableName, object Value) {
            proxyVariables.Add((VariableName, Value));
        }

        private volatile bool _isRunning = false;
        public async Task<dynamic> Run(object Input = null) {

            if (!_isRunning)
            {
                _isRunning = true;
                OnPropertyChanged(nameof(IsRunning));

                Results = new PSDataCollection<PSObject>();

                var rs = RunspaceFactory.CreateRunspace();
                rs.Open();

                dynamic pipeVar = default(dynamic);

                try
                {
                    var _result = new PSDataCollection<PSObject>();

                    if (Input != null) _result.Add((PSObject)Input);

                    bool onFirst = true;

                    currentPowerShell.Runspace = rs;

                    foreach (var s in ScriptFragments)
                    {
                        if (onFirst)
                        {
                            onFirst = false;
                            if (this.Parameters != null)
                            {
                                foreach (var p in Parameters)
                                {
                                    currentPowerShell.AddParameter(p.Key, p.Value);
                                }
                            }
                        }

                        currentPowerShell.Commands.Clear();

                        currentPowerShell.AddScript(s.ToString());
                        currentPowerShell.Runspace.SessionStateProxy.SetVariable("result", pipeVar);

                        // Merge results into a single collection of objects to be parsed later.
                        currentPowerShell.Commands.Commands[0].MergeMyResults(PipelineResultTypes.All, PipelineResultTypes.Output);

                        try
                        {
                            _result = await Task.Factory.FromAsync(currentPowerShell.BeginInvoke<object>(Input), result => currentPowerShell.EndInvoke(result));
                            pipeVar = _result.FirstOrDefault();
                            foreach (var r in _result)
                            {
                                Results?.Add(r);
                            }
                        }
                        catch (Exception e)
                        {
                            Results?.Add(new PSObject(e.GetAllMessages()));
                        }

                        OnPropertyChanged(nameof(Results));
                    }

                    var exit = 0;
                    try { exit = Convert.ToInt32(currentPowerShell.Runspace.SessionStateProxy.GetVariable("LastExitCode")); }
                    catch { /* Don't really care about this */ }
                    ExitCode = exit;

                    OnPropertyChanged(nameof(Results));
                }
                finally
                {
                    currentPowerShell.Runspace = null;
                    rs.Close();
                    IsRunning = false;
                }
            }

            return Results;
        }


#if false
        public void RunAs(AdminUser user, object Input = null)
        {
            try
            {
                var source = Script.ToString().Trim();
                currentPowerShell.AddScript(source, true);

                if (this.Parameters != null)
                {
                    foreach (var p in Parameters)
                    {
                        currentPowerShell.AddParameter(p.Key, p.Value);
                    }
                }

                // Merge results into a single collection of objects to be parsed later.
                currentPowerShell.Commands.Commands[0].MergeMyResults(PipelineResultTypes.All, PipelineResultTypes.Output);

                // If there is any input pass it in, otherwise just invoke
                // the pipeline.
                using (Runspace runspace = RunspaceFactory.CreateRunspace())
                {
                    runspace.Open();

                    currentPowerShell.Runspace = runspace;

                    foreach (var proxyVar in proxyVariables)
                    {
                        currentPowerShell?.Runspace.SessionStateProxy.SetVariable(proxyVar.Item1, proxyVar.Item2);
                    }

                    using (WindowsIdentity newId = new WindowsIdentity(user.IdentityHandle.DangerousGetHandle()))
                    {
                        using (WindowsImpersonationContext impersonatedUser = newId.Impersonate())
                        {
                            Debug.WriteLine($"Env user: {Environment.GetEnvironmentVariable("%username%")}");
                            if (Input != null)
                            {
                                Results = new PSDataCollection<PSObject>(currentPowerShell.Invoke(new object[] { Input }));
                            }
                            else
                            {
                                try
                                {
                                    Results = new PSDataCollection<PSObject>(currentPowerShell.Invoke());
                                }
                                catch (Exception e)
                                {
                                    Results?.Add(new PSObject(e.GetAllMessages()));
                                }
                            }
                        }
                    }

                    var exit = 0;
                    try { exit = Convert.ToInt32(currentPowerShell.Runspace.SessionStateProxy.GetVariable("LastExitCode")); }
                    catch { /* Don't really care about this */ }
                    ExitCode = exit;

                    runspace.Close();
                }
            }
            finally
            {

            }
        }
#endif


        /* Dont use any of this!!!! It doesn't work yet!!!
        private BlockingCollection<InformationalRecord> PSOutput = new BlockingCollection<InformationalRecord>();
        private Task ForwardMessages;
        */


        public async Task<PSDataCollection<PSObject>> RunAsync(object Input = null)
        {
            try
            {
                var source = Script.ToString().Trim();
                currentPowerShell.AddScript(source, true);

                if (this.Parameters != null)
                {
                    foreach (var p in Parameters)
                    {
                        currentPowerShell.AddParameter(p.Key, p.Value);
                    }
                }

                // Merge results into a single collection of objects to be parsed later.
                currentPowerShell.Commands.Commands[0].MergeMyResults(PipelineResultTypes.All, PipelineResultTypes.Output);

                return await Task.Factory.FromAsync(currentPowerShell.BeginInvoke<object>(Input), result => currentPowerShell.EndInvoke(result));
            }
            finally
            {

            }
        }

        public async Task<PSDataCollection<PSObject>> RunAsync(String Source, Dictionary<string, object> Parameters=null, object Input = null)
        {
            try
            {
                currentPowerShell.AddScript(Source, true);

                if (this.Parameters != null)
                {
                    foreach (var p in Parameters)
                    {
                        currentPowerShell.AddParameter(p.Key, p.Value);
                    }
                }

                // Merge results into a single collection of objects to be parsed later.
                currentPowerShell.Commands.Commands[0].MergeMyResults(PipelineResultTypes.All, PipelineResultTypes.Output);

                return await Task.Factory.FromAsync(currentPowerShell.BeginInvoke<object>(Input), result => currentPowerShell.EndInvoke(result));
            }
            finally
            {

            }
        }


        #endregion  // methods
    }
}
